@@ -1,16 +1,15 @@
-use std::{ffi::OsStr, io, path::Path, str::FromStr, sync::Arc, time::Instant};
+use std::{convert::TryFrom, ffi::OsStr, io, path::Path, str::FromStr, sync::Arc, time::Instant};
 
 use anyhow::anyhow;
+use git_repository as git;
 use git_repository::{
     hash,
     hash::ObjectId,
     interrupt,
     objs::bstr::ByteVec,
-    odb::{pack, pack::cache::DecodeEntry, Find},
-    prelude::{Finalize, FindExt},
-    progress,
-    refs::file::ReferenceExt,
-    traverse, Progress,
+    odb::{pack, pack::cache::DecodeEntry, FindExt},
+    prelude::{Finalize, ReferenceAccessExt},
+    progress, traverse, Progress,
 };
 
 use crate::OutputFormat;
@@ -72,12 +71,26 @@ pub struct Context<W> {
     /// If unset, counting will only use one thread and thus yield the same sequence of objects in any case.
     /// If the `thread_limit` is 1, the count is always deterministic.
     pub nondeterministic_count: bool,
+    /// If true, delta objects may refer to their base as reference, allowing it not to be included in the created back.
+    /// Otherwise these have to be recompressed in order to make the pack self-contained.
+    pub thin: bool,
     /// If set, don't use more than this amount of threads.
     /// Otherwise, usually use as many threads as there are logical cores.
     /// A value of 0 is interpreted as no-limit
     pub thread_limit: Option<usize>,
     /// If set, statistics about the operation will be written to the output stream.
     pub statistics: Option<OutputFormat>,
+    /// The size of the cache storing fully decoded delta objects. This can greatly speed up pack decoding by reducing the length of delta
+    /// chains. Note that caches also incur a cost and poorly used caches may reduce overall performance.
+    /// This is a total, shared among all threads if `thread_limit` permits.
+    ///
+    /// If 0, the cache is disabled entirely.
+    pub pack_cache_size_in_bytes: usize,
+    /// The size of the cache to store full objects by their ID, bypassing any lookup in the object database.
+    /// Note that caches also incur a cost and poorly used caches may reduce overall performance.
+    ///
+    /// This is a total, shared among all threads if `thread_limit` permits.
+    pub object_cache_size_in_bytes: usize,
     /// The output stream for use of additional information
     pub out: W,
 }
@@ -91,75 +104,65 @@ pub fn create<W>(
     Context {
         expansion,
         nondeterministic_count,
+        thin,
         thread_limit,
         statistics,
+        pack_cache_size_in_bytes,
+        object_cache_size_in_bytes,
         mut out,
     }: Context<W>,
 ) -> anyhow::Result<()>
 where
     W: std::io::Write,
 {
-    // TODO: review need for Arc for the counting part, use different kinds of Easy depending on need
-    let repo = Arc::new(git_repository::discover(repository_path)?);
+    let repo = git::discover(repository_path)?;
     progress.init(Some(2), progress::steps());
     let tips = tips.into_iter();
     let make_cancellation_err = || anyhow!("Cancelled by user");
-    let (db, input): (
+    let (odb, input): (
         _,
         Box<dyn Iterator<Item = Result<ObjectId, input_iteration::Error>> + Send>,
     ) = match input {
         None => {
+            use git::bstr::ByteSlice;
+            use os_str_bytes::OsStrBytes;
             let mut progress = progress.add_child("traversing");
+            let repo = repo.into_easy();
             progress.init(None, progress::count("commits"));
             let tips = tips
                 .map(|tip| {
-                    ObjectId::from_hex(&Vec::from_os_str_lossy(tip.as_ref())).or_else({
-                        // TODO: Use Easy when ready…
-                        let packed = repo.refs.packed_buffer().ok().flatten();
-                        let refs = &repo.refs;
-                        let repo = Arc::clone(&repo);
-                        move |_| {
-                            refs.find(tip.as_ref().to_string_lossy().as_ref(), packed.as_ref())
-                                .map_err(anyhow::Error::from)
-                                .and_then(|mut r| {
-                                    r.peel_to_id_in_place(refs, packed.as_ref(), |oid, buf| {
-                                        repo.odb
-                                            .try_find(oid, buf, &mut pack::cache::Never)
-                                            .map(|obj| obj.map(|obj| (obj.kind, obj.data)))
-                                    })
-                                    .map(|oid| oid.to_owned())
-                                    .map_err(anyhow::Error::from)
-                                })
-                        }
+                    ObjectId::from_hex(&Vec::from_os_str_lossy(tip.as_ref())).or_else(|_| {
+                        repo.find_reference(tip.as_ref().to_raw_bytes().as_bstr())
+                            .map_err(anyhow::Error::from)
+                            .and_then(|r| r.into_fully_peeled_id().map(|oid| oid.detach()).map_err(Into::into))
                     })
                 })
                 .collect::<Result<Vec<_>, _>>()?;
-            let db = match Arc::try_unwrap(repo) {
-                Ok(repo) => Arc::new(repo.odb),
-                Err(_) => unreachable!("there is only one instance left here"),
-            };
+            let odb = Arc::new(git::Repository::try_from(repo)?.odb);
             let iter = Box::new(
-                // TODO: Easy-based traversal
                 traverse::commit::Ancestors::new(tips, traverse::commit::ancestors::State::default(), {
-                    let db = Arc::clone(&db);
+                    let db = Arc::clone(&odb);
                     move |oid, buf| db.find_commit_iter(oid, buf, &mut pack::cache::Never).ok()
                 })
                 .map(|res| res.map_err(Into::into))
                 .inspect(move |_| progress.inc()),
             );
-            (db, iter)
+            (odb, iter)
         }
         Some(input) => {
-            let db = match Arc::try_unwrap(repo) {
-                Ok(repo) => Arc::new(repo.odb),
-                Err(_) => unreachable!("there is only one instance left here"),
-            };
-            let iter = Box::new(input.lines().map(|hex_id| {
-                hex_id
-                    .map_err(Into::into)
-                    .and_then(|hex_id| ObjectId::from_hex(hex_id.as_bytes()).map_err(Into::into))
-            }));
-            (db, iter)
+            let mut progress = progress.add_child("iterating");
+            progress.init(None, progress::count("objects"));
+            let iter = Box::new(
+                input
+                    .lines()
+                    .map(|hex_id| {
+                        hex_id
+                            .map_err(Into::into)
+                            .and_then(|hex_id| ObjectId::from_hex(hex_id.as_bytes()).map_err(Into::into))
+                    })
+                    .inspect(move |_| progress.inc()),
+            );
+            (Arc::new(repo.odb), iter)
         }
     };
 
@@ -174,21 +177,30 @@ where
         } else {
             Some(1)
         };
-        let make_cache = move || {
-            if may_use_multiple_threads {
+        let (_, _, thread_count) = git::parallel::optimize_chunk_size_and_thread_limit(50, None, thread_limit, None);
+        let make_caches = move || {
+            let per_thread_object_cache_size = object_cache_size_in_bytes / thread_count;
+            let object_cache: Box<dyn pack::cache::Object> = if per_thread_object_cache_size < 10_000 {
+                Box::new(pack::cache::object::Never) as Box<dyn pack::cache::Object>
+            } else {
+                Box::new(pack::cache::object::MemoryCappedHashmap::new(
+                    per_thread_object_cache_size,
+                ))
+            };
+            let per_thread_object_pack_size = pack_cache_size_in_bytes / thread_count;
+            let pack_cache: Box<dyn DecodeEntry> = if per_thread_object_pack_size < 10_000 {
                 Box::new(pack::cache::Never) as Box<dyn DecodeEntry>
             } else {
-                Box::new(pack::cache::lru::MemoryCappedHashmap::new(512 * 1024 * 1024)) as Box<dyn DecodeEntry>
-                // todo: Make that configurable
-            }
+                Box::new(pack::cache::lru::MemoryCappedHashmap::new(per_thread_object_pack_size))
+            };
+            (pack_cache, object_cache)
         };
-        let db = Arc::clone(&db);
         let progress = progress::ThroughputOnDrop::new(progress);
         let input_object_expansion = expansion.into();
         let (mut counts, count_stats) = if may_use_multiple_threads {
             pack::data::output::count::objects(
-                db,
-                make_cache,
+                Arc::clone(&odb),
+                make_caches,
                 input,
                 progress,
                 &interrupt::IS_INTERRUPTED,
@@ -199,9 +211,10 @@ where
                 },
             )?
         } else {
+            let mut caches = make_caches();
             pack::data::output::count::objects_unthreaded(
-                db,
-                &mut make_cache(),
+                Arc::clone(&odb),
+                (&mut caches.0, &mut caches.1),
                 input,
                 progress,
                 &interrupt::IS_INTERRUPTED,
@@ -219,13 +232,13 @@ where
         let progress = progress.add_child("creating entries");
         pack::data::output::InOrderIter::from(pack::data::output::entry::iter_from_counts(
             counts,
-            Arc::clone(&db),
+            Arc::clone(&odb),
             || pack::cache::Never,
             progress,
             pack::data::output::entry::iter_from_counts::Options {
                 thread_limit,
                 mode: pack::data::output::entry::iter_from_counts::Mode::PackCopyAndBaseObjects,
-                allow_thin_pack: false, // todo: make this configurable
+                allow_thin_pack: thin,
                 chunk_size,
                 version: Default::default(),
             },
@@ -313,6 +326,7 @@ fn human_output(
                 decoded_and_recompressed_objects,
                 missing_objects,
                 objects_copied_from_pack,
+                ref_delta_objects,
             },
     }: Statistics,
     mut out: impl std::io::Write,
@@ -333,9 +347,10 @@ fn human_output(
     #[rustfmt::skip]
     writeln!(
         out,
-        "\t{:<width$} {}\n\t{:<width$} {}\n\t{:<width$} {}",
+        "\t{:<width$} {}\n\t{:<width$} {}\n\t{:<width$} {}\n\t{:<width$} {}",
         "decoded and recompressed", decoded_and_recompressed_objects,
         "pack-to-pack copies", objects_copied_from_pack,
+        "ref-delta-objects", ref_delta_objects,
         "missing objects", missing_objects,
         width = width
     )?;
